#include <stdlib.h>

#include "Global.h"
#include "Attack.h"
#include "Creature.h"
#include "Object.h"
#include "Hero.h"

const char *
shAttack::noun ()
{
    switch (mType) {
    case kNoAttack: return "non-attack";
    case kAcidBath: return "acid bath";
    case kAcidSplash: return "acid splash";
    case kAnalProbe: return "probe";
    case kAttach: return "attachment";
    case kBite: return "bite";
    case kBlast: return "explosion";
    case kBullet: return "bullet";
    case kBolt: return "bolt";
    case kBreatheBugs: return "cloud of bugs";
    case kBreatheFire: return "fiery breath";
    case kBreatheTime: return "time warp";
    case kBreatheTraffic: return "packet storm";
    case kBreatheViruses: return "cloud of viruses";
    case kChoke: return "choke";
    case kClaw: return "claw";
    case kClub: return "club";
    case kCombi: return "spearhead";
    case kCook: return "cooker"; /* ugh */
    case kCrush: return "grasp";
    case kCut: return "cut";
    case kDisc: return "spinning edge";
    case kDisintegrationRay: return "disintegration ray";
    case kExplode:
        return mDamage[0].mEnergy == kPsychic ? "psychic wave" : "explosion";
    case kExtractBrain: return "extraction";
    case kFaceHug: return "grasp";
    case kFlash: return "flash";
    case kFreezeRay: return "freeze ray";
    case kGammaRay: return "gamma ray";
    case kGaussRay: return "gauss ray";
    case kHeadButt: return "head butt";
    case kHealingRay: return "healing ray";
    case kHeatRay: return "heat ray";
    case kImpact: return "smash";
    case kIncendiary: return "flame burst";
    case kKick: return "kick";
    case kLightningBolt: return "lightning bolt";
    case kLaser: return "laser beam";
    case kLegalThreat: return "legal threat";
    case kLight: return "blinding flash";
    case kMentalBlast: return "mental blast";
    case kPlasmaGlob: return "plasma glob";
    case kPoisonRay: return "poison ray";
    case kPrick: return "needle";
    case kPsionicStorm: return "psionic storm";
    case kPunch: return "punch";
    case kRestorationRay: return "restoration ray";
    case kQuill: return "quill";
    case kRail: return "rail";
    case kShot: return "shot";
    case kSlash: return "edge";
    case kSlime: return "slime";
    case kSmash: return "smack";
    case kStab: return "stab";
    case kSpear: return "short spear";
    case kSpit: return "spittle";
    case kStasisRay: return "stasis ray";
    case kSting: return "stinger";
    case kSword: return "blade";
    case kTailSlap: return "tail";
    case kTouch: return "touch";
    case kTransporterRay: return "transporter ray";
    case kWeb: return "web";
    case kWhip: return "whip";
    case kVomit: return "vomit";
    case kZap: return "zap";
    case kMaxAttackType: return "program bug";
    }
    return "attack";
}


static const char *
energyFormat (shEnergyType t) {
    switch (t) {
    case kSlashing: return "%d-%d slash";
    case kPiercing: return "%d-%d pierce";
    case kConcussive: return "%d-%d concussion";
    case kBlinding: return "%d-%d blinding";
    case kBrainExtracting: return "neurosurgery";
    case kBugging: return "%d-%d development anti-patterns";
    case kBurning: return "%d-%d fire";
    case kChoking: return "%d-%d oxygen deprivation";
    case kConfusing: return "%d-%d confusing";
    case kCorrosive: return "%d-%d acid";
    case kCreditDraining: return "financial crises";
    case kDigging: return "dig";
    case kDisarming: return "%d-%d disarm";
    case kDisintegrating: return "%d-%d atomization";
    case kElectrical: return "%d-%d shock";
    case kFaceHugging: return "%d-%d implant embryo";
    case kFlare: return "light area";
    case kForce: return "%d-%d force";
    case kFreezing: return "%d-%d ice";
    case kHealing: return "healing";
    case kHosing: return "%d-%d hose";
    case kLaser: return "%d-%d laser";
    case kMagnetic: return "%d-%d gauss";
    case kMesmerizing: return "%d-%d hypnotic";
    case kParalyzing: return "%d-%d paralytic";
    case kPlaguing: return "%d-%d plague";
    case kPoisonous: return "%d-%d poison";
    case kRadiological: return "%d-%d rad";
    case kRestoring: return "restoration";
    case kPsychic: return "%d-%d psi";
    case kSickening: return "%d-%d disease";
    case kStunning: return "%d-%d stun";
    case kTransporting: return "transporting";
    case kTimeWarping: return "%d-%d time warp";
    case kViolating: return "%d-%d violating";
    case kWebbing: return "%d-%d webbing";
    case kNoEnergy: return NULL;
    case kMaxEnergyType: return "program bug";
    }
    return "%d-%d unknown";
}


const char *
energyDescription (shEnergyType t) {
    switch (t) {
    case kSlashing: return "slashing damage";
    case kPiercing: return "piercing damage";
    case kConcussive: return "concussive damage";
    case kBlinding: return "blinding";
    case kBurning: return "fire";
    case kConfusing: return "confusion";
    case kCorrosive: return "acid damage";
    case kElectrical: return "electrical damage";
    case kForce: return "force effects";
    case kLaser: return "laser damage";
    case kFreezing: return "cold damage";
    case kMagnetic: return "magnetic effects";
    case kMesmerizing: return "falling asleep";
    case kParalyzing: return "paralysis";
    case kPlaguing: return "plague";
    case kPoisonous: return "poison";
    case kRadiological: return "radiation";
    case kPsychic: return "psychic attacks";
    case kSickening: return "sickness";
    case kStunning: return "stunning";
    case kViolating: return "violation";

    case kNoEnergy:
    case kBrainExtracting:
    case kBugging:
    case kChoking:
    case kCreditDraining:
    case kDigging:
    case kDisarming:
    case kDisintegrating:
    case kFaceHugging:
    case kFlare:
    case kHealing:
    case kHosing:
    case kRestoring:
    case kTimeWarping:
    case kTransporting:
    case kWebbing:
    case kMaxEnergyType:
        return NULL;
    }
    return NULL;
}

const char *
shAttack::damStr ()
{   /* (2-12 burning + 1-6 gauss), etc. */
    const char *format = energyFormat (mDamage[0].mEnergy);
    if (format) {
        char *buf = GetBuf ();
        int l = 0;
        l += snprintf (buf + l, SHBUFLEN - l, "(");
        for (int i = 0; i < 3 && format; ++i) {
            l += snprintf (buf + l, SHBUFLEN - l, format,
                           mDamage[i].mLow, mDamage[i].mHigh);
            if (i != 2) {
                format = energyFormat (mDamage[i+1].mEnergy);
                if (format) {
                    l += snprintf (buf + l, SHBUFLEN - l, " + ");
                }
            }
        }
        l += snprintf (buf + l, SHBUFLEN - l, ")");
        return buf;
    }
    return "";
}

int
shAttack::getThreatRange (shObject *weapon, shCreature *target)
{
    return weapon ? weapon->getThreatRange (target) : 20;
}

int
shAttack::getCriticalMultiplier (shObject *weapon, shCreature *target)
{
    return weapon ? weapon->getCriticalMultiplier () : 2;
}

int
shAttack::dealsEnergyType (shEnergyType type)
{
    int chance = 0;
    for (int i = 0; i < 3; ++i) {
        if (type == mDamage[i].mEnergy)
            chance = maxi (chance, mDamage[i].mChance);
    }
    return chance;
}

int
shAttack::bypassesShield ()
{
    return mType == kRestorationRay || mType == kHealingRay ||
           mType == kLight || mType == kFlash;
}

int
shAttack::isAbsorbedForFree (int damage)
{
    if (damage == 0) /* Primary damage always costs energy to absorb. */
        return 0;
    return mDamage[damage].mEnergy == kBlinding ||
           mDamage[damage].mEnergy == kParalyzing;
}

int
shAttack::isLightGenerating ()
{ /* Should the attack be drawn over dark squares? */
    return mType == kHeatRay || mType == kIncendiary
        || mType == kLaser   || mType == kLight
        || kLightningBolt;
}

shAttack *newSingleAttack (shAttack::Type type, int flags,
    shEnergyType energy0, int low0, int high0,
    shEnergyType energy1, int low1, int high1,
    shEnergyType energy2, int low2, int high2)
{
    shAttack *atk = new shAttack (kObjNothing, type, shAttack::kSingle, flags,
        FULLTURN, 1, 1, 1,
        energy0, low0, high0, 100,
        energy1, low1, high1, 100,
        energy2, low2, high2, 100);
    return atk;
}

shAttack *newOtherAttack (shAttack::Type type, int flags,
                         shEnergyType energy, int low, int high)
{
    shAttack *atk = new shAttack (kObjNothing, type, shAttack::kOther, flags,
        FULLTURN, 1, 1, 1,
        energy, low, high, 100,
        kNoEnergy, 0, 0, 0,
        kNoEnergy, 0, 0, 0);
    return atk;
}

shAttack *newBurstAttack (shAttack::Type type, int flags, int radius,
                         shEnergyType energy, int low, int high)
{
    shAttack *atk = new shAttack (kObjNothing, type, shAttack::kBurst, flags,
        FULLTURN, 1, 1, radius,
        energy, low, high, 100,
        kNoEnergy, 0, 0, 0,
        kNoEnergy, 0, 0, 0);
    return atk;
}

shAttack *newRayGunAttack (shAttack::Type type,
                          shEnergyType energy, int low, int high)
{
    shAttack *atk = new shAttack (kObjNothing, type, shAttack::kBeam, kAimed,
        FULLTURN, 1, 6, 1,
        energy, low, high, 100,
        kNoEnergy, 0, 0, 0,
        kNoEnergy, 0, 0, 0);
    return atk;
}

shAttack *newMeleeAttack (shAttack::Type type, int time,
    shEnergyType energy0, int low0, int high0,
    shEnergyType energy1, int low1, int high1,
    shEnergyType energy2, int low2, int high2)
{
    return newMeleeAttackP (type, time,
        100, energy0, low0, high0,
        100, energy1, low1, high1,
        100, energy2, low2, high2);
}

shAttack *newMeleeAttackP (shAttack::Type type, int time,
    int prob0, shEnergyType energy0, int low0, int high0,
    int prob1, shEnergyType energy1, int low1, int high1,
    int prob2, shEnergyType energy2, int low2, int high2)
{
    shAttack *atk = new shAttack (kObjNothing, type, shAttack::kSingle,
        kMelee, time, 1, 0, 0,
        energy0, low0, high0, prob0,
        energy1, low1, high1, prob1,
        energy2, low2, high2, prob2);
    return atk;
}

shAttack *newMissileAttack (shAttack::Type type, int time, int range,
    shEnergyType energy0, int low0, int high0,
    shEnergyType energy1, int low1, int high1,
    shEnergyType energy2, int low2, int high2)
{
    shAttack *atk = new shAttack (kObjNothing, type, shAttack::kSingle,
        kMissile, time, 1, range, 0,
        energy0, low0, high0, 100,
        energy1, low1, high1, 100,
        energy2, low2, high2, 100);
    return atk;
}

shAttack *newGrenadeAttack (shAttack::Type type, int radius,
    shEnergyType energy0, int low0, int high0,
    shEnergyType energy1, int low1, int high1,
    shEnergyType energy2, int low2, int high2)
{
    shAttack *atk = new shAttack (kObjNothing, type, shAttack::kBurst,
        kMissile | kTouchAttack, FULLTURN, 1, 10, radius,
        energy0, low0, high0, 100,
        energy1, low1, high1, 100,
        energy2, low2, high2, 100);
    return atk;
}

shAttack *newGunAttack (shAttack::Type type, int time, int range,
    shEnergyType energy0, int low0, int high0,
    shEnergyType energy1, int low1, int high1,
    shEnergyType energy2, int low2, int high2)
{
    return newGunAttackP (type, time, range,
        100, energy0, low0, high0,
        100, energy1, low1, high1,
        100, energy2, low2, high2);
}

shAttack *newGunAttackP (shAttack::Type type, int time, int range,
    int prob0, shEnergyType energy0, int low0, int high0,
    int prob1, shEnergyType energy1, int low1, int high1,
    int prob2, shEnergyType energy2, int low2, int high2)
{
    shAttack *atk = new shAttack (kObjNothing, type, shAttack::kSingle, kAimed,
        time, 1, range, 0,
        energy0, low0, high0, prob0,
        energy1, low1, high1, prob1,
        energy2, low2, high2, prob2);
    return atk;
}

shAttack *newExtendAttack (shAttack::Type type, int time, int range,
    shEnergyType energy0, int low0, int high0,
    shEnergyType energy1, int low1, int high1,
    shEnergyType energy2, int low2, int high2)
{
    shAttack *atk = new shAttack (kObjNothing, type, shAttack::kExtend, kAimed,
        time, 1, range, 0,
        energy0, low0, high0, 100,
        energy1, low1, high1, 100,
        energy2, low2, high2, 100);
    return atk;
}

shAttack *newBeamAttack (shAttack::Type type, int time, int range,
    shEnergyType energy0, int low0, int high0,
    shEnergyType energy1, int low1, int high1,
    shEnergyType energy2, int low2, int high2)
{
    return newBeamAttackP (type, time, range,
        100, energy0, low0, high0,
        100, energy1, low1, high1,
        100, energy2, low2, high2);
}

shAttack *newBeamAttackP (shAttack::Type type, int time, int range,
    int prob0, shEnergyType energy0, int low0, int high0,
    int prob1, shEnergyType energy1, int low1, int high1,
    int prob2, shEnergyType energy2, int low2, int high2)
{
    shAttack *atk = new shAttack (kObjNothing, type, shAttack::kBeam, kAimed,
        time, 1, range, 0,
        energy0, low0, high0, prob0,
        energy1, low1, high1, prob1,
        energy2, low2, high2, prob2);
    return atk;
}

shAttack *newConeAttack (shAttack::Type type, int time, int range, int radius,
    shEnergyType energy0, int low0, int high0,
    shEnergyType energy1, int low1, int high1,
    shEnergyType energy2, int low2, int high2)
{
    shAttack *atk = new shAttack (kObjNothing, type, shAttack::kCone, kAimed,
        time, 1, range, radius,
        energy0, low0, high0, 100,
        energy1, low1, high1, 100,
        energy2, low2, high2, 100);
    return atk;
}

shAttack *newMonsterAttack (shAttack::Type type, int flags, int time, int prob,
    shEnergyType energy0, int low0, int high0,
    shEnergyType energy1, int low1, int high1,
    shEnergyType energy2, int low2, int high2)
{
    shAttack *atk = new shAttack (kObjNothing, type, shAttack::kSingle,
        flags, time, prob, 0, 0,
        energy0, low0, high0, 100,
        energy1, low1, high1, 100,
        energy2, low2, high2, 100);
    return atk;
}

shAttack *newMonsterRangedAttack (shAttack::Type type, shAttack::Effect effect,
    int flags, int time, int prob, int range,
    shEnergyType energy0, int low0, int high0,
    shEnergyType energy1, int low1, int high1,
    shEnergyType energy2, int low2, int high2)
{
    shAttack *atk = new shAttack (kObjNothing, type, effect,
        flags, time, prob, range, 0,
        energy0, low0, high0, 100,
        energy1, low1, high1, 100,
        energy2, low2, high2, 100);
    return atk;
}

shAttack *newAttack (shAttack::Type type, shAttack::Effect effect,
    int flags, int time, int range, int radius,
    shEnergyType energy0, int low0, int high0,
    shEnergyType energy1, int low1, int high1,
    shEnergyType energy2, int low2, int high2)
{
    shAttack *atk = new shAttack (kObjNothing, type, effect,
        flags, time, 1, range, radius,
        energy0, low0, high0, 100,
        energy1, low1, high1, 100,
        energy2, low2, high2, 100);
    return atk;
}

shAttack *ImprovisedObjectAttack =
    newSingleAttack (shAttack::kSmash, kMelee, kConcussive, 1, 3);

shAttack *ImprovisedMissileAttack =
    newSingleAttack (shAttack::kSmash, kMissile, kConcussive, 1, 2);

shAttack *KickedFootballAttack =
    newSingleAttack (shAttack::kSmash, kMissile, kConcussive, 1, 6);

shAttack *GroundCollisionAttack =
    newOtherAttack (shAttack::kSmash, kMissile, kConcussive, 1, 6);

shAttack *PitTrapDamage =
    newSingleAttack (shAttack::kSmash, 0, kConcussive, 1, 6);

shAttack *AcidPitTrapDamage =
    newSingleAttack (shAttack::kAcidBath, 0, kCorrosive, 1, 4);

shAttack *AcidPitSipDamage =
    newOtherAttack (shAttack::kZap, 0, kCorrosive, 3, 8);

shAttack *AcidBloodAttack =
    newGrenadeAttack (shAttack::kAcidSplash, 1, kCorrosive, 2, 12);

shAttack *ExplodingMonsterAttack =
    newBurstAttack (shAttack::kBlast, 0, 1, kConcussive, 6, 18);

shAttack *FallingVatAttack =
    newSingleAttack (shAttack::kSmash, 0, kConcussive, 10, 30);

shAttack *RestrainingBoltWear =
    newSingleAttack (shAttack::kZap, 0, kBugging, 1, 1);

shAttack *TorcRemoveAttack =
    newOtherAttack (shAttack::kZap, 0, kElectrical, 4, 9);

shAttack *GloryDeviceAttack[5] =
{
    /* Probably you had buggy one stuck from trying unided stuff. */
    newBurstAttack (shAttack::kBlast, 0, 2, kConcussive, 6, 18),
    /* Second one likely got by a typo or other stupid mistake. */
    newBurstAttack (shAttack::kBlast, 0, 4, kConcussive, 8, 24),
    /* Did you intentionally install that many bombs? */
    newBurstAttack (shAttack::kBlast, 0, 6, kConcussive, 12, 36),
    /* Someone is doing batshit crazy stuff! */
    newBurstAttack (shAttack::kBlast, 0, 8, kConcussive, 16, 48),
    /* Why you would anyone try this?  Probably only for kicks.
       So, as a reward you get to blow up one level to smithereens. */
    newBurstAttack (shAttack::kBlast, 0, 24, kDisintegrating, 500, 500)
};

/* Where else could it belong? */
void
shHero::quaffFromAcidPit ()
{
    if (Hero.sufferDamage (AcidPitSipDamage)) {
        Hero.die (kKilled, "drinking from an acid pit");
    }
    Hero.abortion ();
}
