#include "Global.h"
#include "Util.h"
#include "Hero.h"
#include "Object.h"
#include "Interface.h"

static int
useHealthMonitor (shObject *monitor)
{
    if (monitor->isBuggy ()) {
        I->p ("Firmware malfunction.  Unable to gauge radiation level.");
        monitor->setBugginessKnown ();
    } else {
        I->p ("Reported radiation dosage absorbed by your body is %s.",
            Hero.radLevels ());
    }
    return 0;
}

static int
useGloryDevice (shObject *device)
{
    if (I->yn ("Activate the glory device and die a martyr's death?")) {
        Hero.die (kSuicide); /* Death routine handles everything else. */
    }
    return 0;
}

void
initializeImplants ()
{
    /* enum: kObjAnyImplant */
    /* enum: kObjFirstImplant = kObjAnyImplant */
    NEW_IMPLANT(kObjAnyImplant,
        "bionic implant", "bionic implant",
        "bionic implant", kGray,
        kObjNothing, 0, shObjectIlk::kAnyBrain,
        kPreidentified, kNoSkillCode, 0, 0, ABSTRACT);

    /* enum: kObjBeneficialImplant */
    NEW_IMPLANT(kObjBeneficialImplant,
        "beneficial implant", "bionic implant",
        "bionic implant", kGray,
        kObjAnyImplant, 0, shObjectIlk::kAnyBrain,
        kPreidentified, kNoSkillCode, 0, 0, ABSTRACT);

    /* enum: kObjHarmfulImplant */
    NEW_IMPLANT(kObjHarmfulImplant,
        "harmful implant", "bionic implant",
        "bionic implant", kGray,
        kObjAnyImplant, 0, shObjectIlk::kAnyBrain,
        kPreidentified, kNoSkillCode, 0, 0, ABSTRACT);

    /* enum: kObjHealthMonitor */
    int n = 0;
    NEW_IMPLANT(kObjHealthMonitor,
        "health monitor", "bionic implant",
        ImplantData[n].mDesc, ImplantData[n].mColor,
        kObjBeneficialImplant, 0, shObjectIlk::kAnyBrain,
        kHealthMonitoring, kNoSkillCode, -3, 300 BZ, 50);
    AllIlks[kObjHealthMonitor]->mUseFunc = useHealthMonitor;

    /* enum: kObjRadiationProcessor */
    ++n;
    NEW_IMPLANT(kObjRadiationProcessor,
        "radiation processor", "bionic implant",
        ImplantData[n].mDesc, ImplantData[n].mColor,
        kObjBeneficialImplant, 0, shObjectIlk::kAnyBrain,
        kRadiationProcessing, kNoSkillCode, -3, 300 BZ, 50);

    /* enum: kObjPoisonResistor */
    ++n;
    /* +10 poison resistance */
    NEW_IMPLANT(kObjPoisonResistor,
        "poison resistor", "bionic implant",
        ImplantData[n].mDesc, ImplantData[n].mColor,
        kObjBeneficialImplant, 0, shObjectIlk::kAnyBrain,
        0, kNoSkillCode, -3, 300 BZ, 50);

    /* enum: kObjBabelFish */
    NEW_IMPLANT(kObjBabelFish,
        "babel fish", "fish", "yellow fish", kYellow,
        kObjBeneficialImplant, 0, shObjectIlk::kAnyEar,
        kTranslation, kNoSkillCode, 0, 300 BZ, 50);

    /* enum: kObjCerebralCoprocessor */
    ++n;
    /* grants enhancement to intelligence */
    NEW_IMPLANT(kObjCerebralCoprocessor,
        "cerebral coprocessor", "bionic implant",
        ImplantData[n].mDesc, ImplantData[n].mColor,
        kObjBeneficialImplant, kEnhanceable, shObjectIlk::kAnyBrain,
        0, kNoSkillCode, -3, 300 BZ, 50);

    /* enum: kObjReflexCoordinator */
    ++n;
    /* grants enhancement to agility */
    NEW_IMPLANT(kObjReflexCoordinator,
        "reflex coordinator", "bionic implant",
        ImplantData[n].mDesc, ImplantData[n].mColor,
        kObjBeneficialImplant, kEnhanceable, shObjectIlk::kAnyBrain,
        0, kNoSkillCode, -3, 300 BZ, 50);

    /* enum: kObjAdrenalineGenerator */
    ++n;
    /* grants enhancement to strength */
    NEW_IMPLANT(kObjAdrenalineGenerator,
        "adrenaline generator", "bionic implant",
        ImplantData[n].mDesc, ImplantData[n].mColor,
        kObjBeneficialImplant, kEnhanceable, shObjectIlk::kAnyBrain,
        0, kNoSkillCode, -3, 300 BZ, 50);

    /* enum: kObjNarcoleptor */
    ++n;
    NEW_IMPLANT(kObjNarcoleptor,
        "narcoleptor", "bionic implant",
        ImplantData[n].mDesc, ImplantData[n].mColor,
        kObjHarmfulImplant, kUsuallyBuggy, shObjectIlk::kAnyBrain,
        kNarcolepsy, kNoSkillCode, -3, 400 BZ, 50);

    /* enum: kObjGloryDevice */
    ++n;
    NEW_IMPLANT(kObjGloryDevice,
        "glory device", "bionic implant",
        ImplantData[n].mDesc, ImplantData[n].mColor,
        kObjHarmfulImplant, kUsuallyBuggy, shObjectIlk::kAnyBrain,
        0, kNoSkillCode, -3, 400 BZ, 50);
    AllIlks[kObjGloryDevice]->mUseFunc = useGloryDevice;

    /* enum: kObjTissueRegenerator */
    ++n;
    /* regenerates hit points */
    NEW_IMPLANT(kObjTissueRegenerator,
        "tissue regenerator", "bionic implant",
        ImplantData[n].mDesc, ImplantData[n].mColor,
        kObjBeneficialImplant, 0, shObjectIlk::kAnyBrain,
        kAutoRegeneration, kNoSkillCode, -3, 400 BZ, 50);

    /* enum: kObjPsiAmp */
    ++n;
    NEW_IMPLANT(kObjPsiAmp,
        "psionic amplifier", "bionic implant",
        ImplantData[n].mDesc, ImplantData[n].mColor,
        kObjBeneficialImplant, kEnhanceable, shObjectIlk::kAnyBrain,
        0, kNoSkillCode, 0, 400 BZ, 50);

    /* enum: kObjMechaDendrites */
    NEW_IMPLANT(kObjMechaDendrites,
        "set of mecha-dendrites", "set of mecha-dendrites",
        "set of mecha-dendrites", kBrightBlue,
        kObjAnyImplant, kPreidentified, shObjectIlk::kCerebellum,
        0, kNoSkillCode, -5, 700 BZ, 40);
    AllIlks[kObjMechaDendrites]->mUseFunc = makeRepair;
    AllIlks[kObjMechaDendrites]->mMaterial = kSteel;

    /* enum: kObjTorc */
    NEW_IMPLANT(kObjTorc,
        "torc", "torc", "torc", kGray,
        kObjAnyImplant, kUsuallyBuggy | kPreidentified, shObjectIlk::kNeck,
        kTelepathy, kNoSkillCode, 0, 250 BZ, 25);
    AllIlks[kObjTorc]->mMaterial = kTitanium;


    /* enum: kObjGenericSkillsoft */
    NEW_IMPLANT(kObjGenericSkillsoft,
        "skillsoft", "bionic implant",
        "bionic implant", kGray,
        kObjBeneficialImplant, 0, shObjectIlk::kAnyBrain,
        kPreidentified, kNoSkillCode, -2, 0, ABSTRACT);

    /* enum: kObjSearchSkillsoft */
    n++;
    NEW_IMPLANT(kObjSearchSkillsoft,
        "search skillsoft", "bionic implant",
        ImplantData[n].mDesc, ImplantData[n++].mColor,
        kObjGenericSkillsoft, 0, shObjectIlk::kAnyBrain,
        kAutoSearching, kSearch, -2, 200 BZ, 50);

    /* enum: kObjRepairSkillsoft */
    n++;
    NEW_IMPLANT(kObjRepairSkillsoft,
        "repair skillsoft", "bionic implant",
        ImplantData[n].mDesc, ImplantData[n++].mColor,
        kObjGenericSkillsoft, 0, shObjectIlk::kAnyBrain,
        0, kRepair, -2, 200 BZ, 50);

    /* enum: kObjBOFHEye */
    /* enum: kObjLastImplant = kObjBOFHEye */
    NEW_IMPLANT(kObjBOFHEye,
        "Eye of the BOFH", "detached retina", "detached retina",
        kRed, kObjNothing, kUnique | kIndestructible, shObjectIlk::kAnyEye,
        kNightVision, kNoSkillCode, 0, 1000 BZ, 0);
}


const char *
describeImplantSite (shObjectIlk::Site site)
{
    switch (site) {
    case shObjectIlk::kFrontalLobe: return "frontal lobe";
    case shObjectIlk::kTemporalLobe: return "temporal lobe";
    case shObjectIlk::kParietalLobe: return "parietal lobe";
    case shObjectIlk::kOccipitalLobe: return "occipital lobe";
    case shObjectIlk::kCerebellum: return "cerebellum";
    case shObjectIlk::kLeftEar: return "left ear";
    case shObjectIlk::kRightEyeball: return "right eyeball";
    case shObjectIlk::kNeck: return "neck";
    default:
        return "brain";
    }
}
